package com.yizhi.xxl.job.executor.service.job;

import cn.hutool.core.date.DateUtil;
import com.xxl.job.core.log.XxlJobLogger;
import com.yizhi.course.application.feign.RemoteDaIndicatorCourseClient;
import com.yizhi.system.application.system.remote.RemoteDaIndicatorSystemClient;
import com.yizhi.system.application.system.remote.SiteClient;
import com.yizhi.system.application.vo.RemoteDaIndicatorVo;
import com.yizhi.training.application.feign.RemoteDaIndicatorTpClient;
import org.apache.commons.collections.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;
import java.util.List;

/**
 * 首页dashboard指标job
 */
@Component
public class DashboardJob implements BaseJob {

    private static final Logger log = LoggerFactory.getLogger(DashboardJob.class);

    @Autowired
    private RemoteDaIndicatorSystemClient systemClient;
    @Autowired
    private RemoteDaIndicatorCourseClient courseClient;
    @Autowired
    private RemoteDaIndicatorTpClient tpClient;
    @Autowired
    private SiteClient siteClient;

    @Override
    public void execute() throws Exception {
        List<Long> allSiteIds = siteClient.getAllSiteId();
        if (CollectionUtils.isEmpty(allSiteIds)) {
            XxlJobLogger.log("-------------------- dashboard 未查询到站点列表，结束执行 ----------------------");
            return;
        }

        Date now = new Date();
        Date yesterday = DateUtil.offsetDay(now, -1).toJdkDate();
        Date endOfWeek = DateUtil.beginOfDay(DateUtil.endOfWeek(yesterday)).toJdkDate();
        Date startOfWeek = DateUtil.beginOfDay(DateUtil.beginOfWeek(yesterday)).toJdkDate();

        for (Long siteId : allSiteIds) {
            try {
                XxlJobLogger.log("-------------------- dashboard 折线图 开始处理 site_id: {} ----------------------", siteId);
                systemClient.accountDay(new RemoteDaIndicatorVo(siteId, null, null, yesterday, now));
                XxlJobLogger.log("-------------------- dashboard 折线图 处理完成 site_id: {} ----------------------", siteId);
            } catch (Exception e) {
                e.printStackTrace();
                throw e;
            }

            // 如果是最后一天，则统计dashboard4个指标
            if (DateUtil.isSameDay(yesterday, endOfWeek)) {
                try {
                    XxlJobLogger.log("-------------------- dashboard 活跃用户比例 开始处理 site_id: {} ----------------------", siteId);
                    systemClient.account(new RemoteDaIndicatorVo(siteId, startOfWeek, endOfWeek, null, now));
                    XxlJobLogger.log("-------------------- dashboard 活跃用户比例 处理完成 site_id: {} ----------------------", siteId);
                } catch (Exception e) {
                    e.printStackTrace();
                    throw e;
                }

                try {
                    XxlJobLogger.log("-------------------- dashboard 课程完成数 开始处理 site_id: {} ----------------------", siteId);
                    courseClient.courseFinish(new com.yizhi.course.application.vo.RemoteDaIndicatorVo(siteId, startOfWeek, endOfWeek, null, now));
                    XxlJobLogger.log("-------------------- dashboard 课程完成数 处理完成 site_id: {} ----------------------", siteId);
                } catch (Exception e) {
                    e.printStackTrace();
                    throw e;
                }

                try {
                    XxlJobLogger.log("-------------------- dashboard 课程学习时长 开始处理 site_id: {} ----------------------", siteId);
                    courseClient.courseTime(new com.yizhi.course.application.vo.RemoteDaIndicatorVo(siteId, startOfWeek, endOfWeek, null, now));
                    XxlJobLogger.log("-------------------- dashboard 课程学习时长 处理完成 site_id: {} ----------------------", siteId);
                } catch (Exception e) {
                    e.printStackTrace();
                    throw e;
                }

                try {
                    XxlJobLogger.log("-------------------- dashboard 项目通过率 开始处理 site_id: {} ----------------------", siteId);
                    tpClient.tpFinish(new com.yizhi.training.application.vo.RemoteDaIndicatorVo(siteId, startOfWeek, endOfWeek, null, now));
                    XxlJobLogger.log("-------------------- dashboard 项目通过率 处理完成 site_id: {} ----------------------", siteId);
                } catch (Exception e) {
                    e.printStackTrace();
                    throw e;
                }
            }
        }
    }

}
