/**
 * FileName: HttpSetjob
 * Author:   wenjunlong
 * Date:     2018/5/26 13:59
 * Description: 发送请求job
 * History:
 * <author>          <time>          <version>          <desc>
 * 作者姓名           修改时间           版本号              描述
 */
package com.yizhi.xxl.job.executor.service.job;

import com.alibaba.fastjson.JSON;
import com.xxl.job.core.log.XxlJobLogger;
import com.yizhi.application.feign.StudyLogClient;
import com.yizhi.application.vo.StudyLogVo;
import com.yizhi.core.application.cache.RedisCache;
import com.yizhi.core.application.cache.distributedlock.impl.RedisDistributedLock;
import com.yizhi.course.application.feign.CourseClient;
import com.yizhi.course.application.feign.CourseStudyClient;
import com.yizhi.course.application.feign.RecordeClient;
import com.yizhi.course.application.vo.HeartBeatVo;
import com.yizhi.course.application.vo.StudyLog;
import com.yizhi.course.application.vo.domain.ChapterEntityVo;
import com.yizhi.course.application.vo.domain.CourseEntityVo;
import com.yizhi.course.application.vo.domain.MaterialEntityVo;
import com.yizhi.course.application.vo.domain.RecordeEntityVo;
import com.yizhi.util.application.date.DateUtil;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.*;
import java.util.stream.Collectors;

/**
 * 课程学习心跳处理
 * <br> 原处理已经备份到HttpSetjobBack
 *
 * @author 谢海军
 * @create 2019/1/23
 * @since 1.0.0
 */
@Component
public class HttpSetjob implements BaseJob {


    private final static String REDIS_KEY = "httpSetjob-singleton";

    private final static String SPILT = ",";

    private final static String REDIS_KEY_MAX = "heartbeat.max.id";

    private final static String REDIS_RERTY_MAX = "heartbeat.retry.max";

    private final static String COURSE_DISTRIBUTED_LOCK = "distributed.lock:course";

    @Autowired
    private StudyLogClient studyLogClient;

    @Autowired
    private RecordeClient recordeClient;

    @Autowired
    private CourseStudyClient courseStudyClient;

    @Autowired
    private RedisDistributedLock redisDistributedLock;

    @Autowired
    private RedisCache redisCache;

    @Autowired
    private CourseClient courseClient;

    @Override
    public void execute() throws Exception {

        String uuid = UUID.randomUUID().toString();
        // 四分分钟的锁等待（重试次数乘等待时间 < 过期时间） 加锁不放在try，免得finally块执行解锁
        //boolean lockStatus = redisDistributedLock.lock(REDIS_KEY,250000,24,10000);
        boolean lockStatus = false;
        Date execDate = Calendar.getInstance().getTime();
        //XxlJobLogger.log(uuid+"批次,心跳处理开始时间{}，分布式锁的状态={}", DateUtil.format(execDate, "yyyy-MM-dd HH:mm:ss.SSS"),lockStatus);
        /*if(lockStatus){
            XxlJobLogger.log(uuid+"批次,分布式锁的状态={},加锁成功，业务开始",lockStatus);
        }else{
            XxlJobLogger.log(uuid+"批次,分布式锁的状态={},加锁失败，业务不进行处理",lockStatus);
            return;
        }*/
        try {
            Map<String, List<StudyLogVo>> errorData = new HashMap<>();

            // 获取最大的id
            Object maxIdObj = redisCache.get(REDIS_KEY_MAX);
            XxlJobLogger.log(uuid + "批次,redis存在的最大id={}", maxIdObj);
            if (Objects.nonNull(maxIdObj) && "off".equalsIgnoreCase(maxIdObj.toString())) {
                XxlJobLogger.log(uuid + "批次,心跳服务开关为关闭，停止心跳处理");
                return;
            }

            // 获取最大的id
            Object retryMaxRedisObj = redisCache.get(REDIS_RERTY_MAX);
            int retryMaxRedis = 1;
            if (Objects.nonNull(retryMaxRedisObj)) {
                retryMaxRedis = Integer.valueOf(retryMaxRedisObj.toString());
            }
            XxlJobLogger.log(uuid + "批次,redis重试次数={}", retryMaxRedisObj);

            List<StudyLogVo> studyLogs = studyLogClient.list("default");

            //XxlJobLogger.log("原始数据={}",JSON.toJSONString(studyLogs));
            if (CollectionUtils.isNotEmpty(studyLogs)) {

                if (Objects.nonNull(maxIdObj)) {
                    long maxIdTmp = Long.valueOf(maxIdObj.toString());
                    // 获取数据库最小的id
                    //Long minId = studyLogClient.minId();
                    Long minId = studyLogs.get(0).getId();
                    XxlJobLogger.log(uuid + "批次,处理的数据库的最小id={},maxIdTmp={}", minId, maxIdTmp);
                    //if (minId <= maxIdTmp && retryMaxRedis<2) {
                    if (minId <= maxIdTmp) {
                        retryMaxRedis = retryMaxRedis + 1;
                        if (retryMaxRedis >= 35) {
                            XxlJobLogger.log(uuid + "！！！！！！！！！！！！！！批次已经重试{}次，现在进行重置归零操作##########################", 30);
                            redisCache.set(REDIS_KEY_MAX, "0");
                            retryMaxRedis = 1;
                        }
                        redisCache.set(REDIS_RERTY_MAX, retryMaxRedis + "");
                        XxlJobLogger.log(uuid + "批次,新的批处理本次查询的数据库最小id={}小于上次redis存放的最大id={}，说明上次处理的还没处理完成，结束本次处理", minId, maxIdTmp);
                        return;
                    }
                }

                HeartBeatVo heartBeatVo = new HeartBeatVo();// todo 心跳参数设置，目前为vo固定，获取根据接口获取

                // 查询scorm课程的心跳记录
                List<StudyLogVo> scormStudyLogs = studyLogClient.list("scorm");
                if (CollectionUtils.isNotEmpty(scormStudyLogs)) {
                    Map<String, List<StudyLogVo>> scormStudyData = scormStudyLogs.parallelStream()
                            .collect(Collectors.groupingBy(scormObjKey -> scormObjKey.getAccountId() + "#" + scormObjKey.getCourseId()));
                    for (Map.Entry<String, List<StudyLogVo>> scormItem : scormStudyData.entrySet()) {
                        //Optional<StudyLog> validData = scormItem.getValue().parallelStream().filter(scormItemFilter -> scormItemFilter.getChapterId()>0).findAny();
                        List<StudyLogVo> dataSort = scormItem.getValue().stream().sorted(Comparator.comparing(StudyLogVo::getNginxTime)).collect(Collectors.toList());
                        int size = dataSort.size();
                        StudyLogVo studyLogScorm = dataSort.get(size - 1);
                        studyLogScorm.setTerminalDevice((size * heartBeatVo.getCount()) + "");
                        studyLogScorm.setCreateTime(dataSort.get(0).getNginxTime());
                        StudyLog courseStudentLogVo = new StudyLog();
                        BeanUtils.copyProperties(studyLogScorm,courseStudentLogVo);
                        courseClient.saveScormHeartBeat(courseStudentLogVo);
                        /*if(validData.isPresent()){
                            // 传输数据进行更新
                            StudyLog studyLogScorm = validData.get();
                            List<StudyLog> data = scormItem.getValue();

                            studyLogScorm.setTerminalDevice((scormItem.getValue().size()*heartBeatVo.getCount())+"");
                            courseClient.saveScormHeartBeat(studyLogScorm);
                        }*/
                    }
                    boolean flag = studyLogClient.deleteBatch(scormStudyLogs.stream().map(obj -> obj.getId()).collect(Collectors.toList()));
                    XxlJobLogger.log("scrom课程的时长处理完毕");
                } else {
                    XxlJobLogger.log("没有scrom心跳信息");
                }

                Long maxId = studyLogs.get(studyLogs.size() - 1).getId();
                XxlJobLogger.log(uuid + "批次,处理的数据库的最大id={}", maxId);
                redisCache.set(REDIS_KEY_MAX, maxId.toString());
                redisCache.set(REDIS_RERTY_MAX, "1");

                XxlJobLogger.log("查询到studyLog数据{}条", studyLogs.size());

                //studyLogs.stream().parallel().collect(Collectors.groupingBy(obj->obj.getSessionId(),HashMap::new,Collectors.toCollection(LinkedList::new)));
                Map<String, List<StudyLogVo>> data = studyLogs.stream().parallel().collect(Collectors.groupingBy(
                        //obj -> obj.getSessionId() + SPILT + obj.getCourseId() + SPILT + obj.getChapterId()));//,Collectors.counting()
                        obj -> obj.getSessionId()));
                XxlJobLogger.log("分组SessionId以后的数据为{}条", data.size());
                int excutIndex = 0;
                //Set<String> keySets = data.keySet();
                List<String> sessionIds = new ArrayList<>();
                Set<Long> courseIds = new HashSet<>();
                List<Long> chapterIds = new ArrayList<>();
                //for (String keyVal : keySets) {
                for (Map.Entry<String, List<StudyLogVo>> entryObj : data.entrySet()) {

                    //String[] valTemp = keyVal.split(SPILT);
                    sessionIds.add(entryObj.getKey());
                    List<StudyLogVo> studyLogsIn = entryObj.getValue();
                    StudyLogVo studyLogIn = studyLogsIn.get(0);
                    Long courseIdIn = studyLogIn.getCourseId();
                    Long chapterId = studyLogIn.getChapterId();
                    if (Objects.isNull(courseIdIn) || courseIdIn.longValue() <= 0) {
                        StudyLogVo studyLogIn1 = studyLogsIn.stream().filter(obj -> (Objects.nonNull(obj.getChapterId()) && obj.getChapterId().longValue() > 0))
                                .findFirst().get();
                        courseIdIn = studyLogIn1.getCourseId();
                        chapterId = studyLogIn1.getChapterId();

                    }
                    courseIds.add(courseIdIn);
                    chapterIds.add(chapterId);
                }
                // 查询已经存在的学习记录
                // XxlJobLogger.log("sessionIds={}", sessionIds);
                List<RecordeEntityVo> recordes = recordeClient.getExistRecorde(sessionIds);

                //XxlJobLogger.log("根据sessionIds={}，查询结果={}",JSON.toJSONString(sessionIds),JSON.toJSONString(recordes));
                Map<String, RecordeEntityVo> recordeMap = null; // 存在的学习记录
                boolean existFlag = false;
                if (CollectionUtils.isNotEmpty(recordes)) {
                    existFlag = true;
                    try {
                        recordeMap = recordes.stream().collect(Collectors.toMap(key -> key.getSessionId(), val -> val));
                    } catch (Exception e) {
                        XxlJobLogger.log("存在重复的学习课程，进行特殊处理，数据={}", JSON.toJSONString(recordes));
                        recordeMap = new HashMap<>();
                        Map<String, List<RecordeEntityVo>> recordesMap = recordes.stream().collect(Collectors.groupingBy(key -> key.getSessionId()));
                        for (Map.Entry<String, List<RecordeEntityVo>> item : recordesMap.entrySet()) {
                            List<RecordeEntityVo> recordeListT = item.getValue();
                            int legt = recordeListT.size();
                            if (legt == 1) {
                                recordeMap.put(item.getKey(), recordeListT.get(0));
                            } else {
                                recordeListT.sort(new Comparator<RecordeEntityVo>() {
                                    @Override
                                    public int compare(RecordeEntityVo o1, RecordeEntityVo o2) {
                                        return o2.getEndTime().compareTo(o1.getEndTime());
                                    }
                                });
                                recordeMap.put(item.getKey(), recordeListT.get(legt - 1));
                            }
                        }
                    }
                }
                //XxlJobLogger.log("根据sessionIds={}，查询结果转成map={}",JSON.toJSONString(sessionIds),JSON.toJSONString(recordeMap));
                // 查询课程信息
                List<CourseEntityVo> courseList = null;
                try {

                    if (recordeMap != null) {
                        for (RecordeEntityVo recordeItems : recordeMap.values()) {
                            courseIds.add(recordeItems.getCourseId());
                            chapterIds.add(recordeItems.getChapterId());
                        }
                    }
                    courseList = recordeClient.selectExistCourse(courseIds);
                } catch (Exception e) {
                    return;
                }
                Map<Long, CourseEntityVo> courseMap = courseList.stream().collect(Collectors.toMap(key -> key.getId(), val -> val));

                // 查询章节信息
                List<ChapterEntityVo> chapterList = recordeClient.selectExistChapter(chapterIds);
                Map<Long, ChapterEntityVo> chapterMap = chapterList.stream().collect(Collectors.toMap(key -> key.getId(), val -> val));

                // 查询素材信息
                List<Long> materials = chapterList.stream().map(obj -> obj.getMaterialId()).collect(Collectors.toList());
                List<MaterialEntityVo> materialList = recordeClient.selectExistMaterial(materials);
                Map<Long, MaterialEntityVo> materialMap = materialList.stream().collect(Collectors.toMap(key -> key.getId(), val -> val));

                // 处理心跳数据
                List<RecordeEntityVo> newRecorde = new ArrayList<>();
                //List<Recorde> existRecorde = new ArrayList<>();
                // 更新的学习记录
                List<RecordeEntityVo> upRecorde = new ArrayList<>();

                //XxlJobLogger.log("对原始数据进行分组以后的数据={}",JSON.toJSONString(data));
                for (Map.Entry<String, List<StudyLogVo>> item : data.entrySet()) {
                    excutIndex++;
                    XxlJobLogger.log("分组SessionId以后的开始处理{}条数据", excutIndex);
                    String keyVal = item.getKey();
                    String[] keyValArrays = keyVal.split(SPILT);
                    String sessionId = keyValArrays[0];
                    if (StringUtils.isBlank(sessionId)) {
                        //XxlJobLogger.log("sessionId为空={}", keyVal);
                        List<StudyLogVo> studyLogVoList = item.getValue();
                        List<StudyLogVo> courseStudyLogVOList = getCourseStudyLogVO(studyLogVoList);
                        errorData.put(UUID.randomUUID().toString() + ":   sessionId为空", courseStudyLogVOList);
                        XxlJobLogger.log("sessionId为空的数据={}", JSON.toJSONString(item));
                        continue;
                    }
                    List<StudyLogVo> valList = item.getValue();
                    // 每次处理的批次的数据条数
                    int studyLogSize = valList.size();
                    // 学习时间次数（对相同时间的同一个sessionId进行合并）
                    //int groupSize = valList.stream().collect(Collectors.groupingBy(gkey -> gkey.getNginxTime().getTime())).size();
                    //sortList(valList);
                    // 常规逆序 numList.stream().sorted();(常规升序)
                    //numList = numList.stream().sorted(Comparator.reverseOrder()).collect(Collectors.toList());
                    // 对象排序 applelist.stream().sorted(Comparator.comparing(Apple::getWight).reversed())；（逆序）
                    //List<Apple> applelist1 = applelist.stream().sorted(Comparator.comparing(Apple::getWight)).collect(Collectors.toList());
                    /*Collections.sort(valList, (e1, e2) -> {
                        //return e1.getWight() - e2.getWight();
                        return e1.getNginxTime().compareTo(e2.getNginxTime());
                    });*/
                    valList = valList.stream().sorted(Comparator.comparing(StudyLogVo::getNginxTime)).collect(Collectors.toList());

                    //XxlJobLogger.log("排序后的数据={}",JSON.toJSONString(valList));
                    // 合并心跳记录转换成学习记录
                    RecordeEntityVo recorde = new RecordeEntityVo();
                    StudyLogVo lastStudy = valList.get(studyLogSize - 1);
                    // 获取最后的学习时间
                    //long lastStudyDateVal = lastStudy.getTimestamp();
                    //Date lastStudyDate = new Date(lastStudyDateVal);

                    Integer progress = lastStudy.getProgress();

                    Date lastStudyDate = lastStudy.getNginxTime();
                    //XxlJobLogger.log("最后学习时间={}",lastStudyDate);

                    StudyLogVo startStudyLog = valList.get(0);

                    // 获取开始时间
                    //long startStudyDateVal = startStudyLog.getTimestamp();
                    Date firstStartDate = startStudyLog.getNginxTime();

                    // 获取课程的策略
                    Long courseIdBycurrStudy = startStudyLog.getCourseId();
                    CourseEntityVo course = courseMap.get(courseIdBycurrStudy);
                    if (course == null) {
                        if (existFlag && recordeMap.containsKey(sessionId)) {
                            courseIdBycurrStudy = recordeMap.get(sessionId).getCourseId();
                            course = courseMap.get(courseIdBycurrStudy);
                        }
                        if (course == null) {
                            //XxlJobLogger.log("课程id={}的心跳的课程不存在", startStudyLog.getCourseId());
                            List<StudyLogVo> studyLogVoList = item.getValue();
                            List<StudyLogVo> courseStudyLogVOList = getCourseStudyLogVO(studyLogVoList);
                            errorData.put("课程id=" + startStudyLog.getCourseId() + "的心跳的课程不存在", courseStudyLogVOList);
                            XxlJobLogger.log("课程id为空的数据={}", JSON.toJSONString(item));
                            continue;
                        }
                    }

                    int duration = 3;
                    int overed = 0;
                    String extedText = "心跳未完成";
                    Integer overedTmp = lastStudy.getMaterialCompleted();
                    if (Integer.valueOf(1).equals(overedTmp)) {
                        extedText = "顺时间心跳完成";
                        overed = 1;
                    } else {
                        Optional<StudyLogVo> dataOpt = valList.stream().parallel().filter(itemStudy -> Integer.valueOf(1).equals(itemStudy.getMaterialCompleted())).findAny();
                        if (dataOpt.isPresent()) {
                            //StudyLog headObj = dataOpt.get();
                            overed = 1;
                            extedText = "非顺序心跳完成";
                        }
                    }


                    RecordeEntityVo existRecordeItem = null; // 是否是有学习过的记录
                    if (existFlag && recordeMap.containsKey(sessionId)) {
                        // 获取存在的学习记录
                        existRecordeItem = recordeMap.get(sessionId);
                        duration = existRecordeItem.getDuration();// 获取以前学习时长

                        // 组装学习记录对象
                        BeanUtils.copyProperties(existRecordeItem, recorde);

                        String terminalDevice = existRecordeItem.getTerminalDevice();
                        if (terminalDevice.indexOf(extedText) < 0) {
                            terminalDevice += "-->" + startStudyLog.getTerminalDevice() + "[" + DateUtil.toShortSeconds(new Date()) + "]" + recorde.getOvered() + extedText;
                            recorde.setTerminalDevice(terminalDevice);
                        } else if (overed == 1) {
                            terminalDevice += "-->" + "[" + DateUtil.toShortSeconds(new Date()) + "]" + recorde.getOvered() + extedText;
                            recorde.setTerminalDevice(terminalDevice);
                        }
                        String startTerminalType = existRecordeItem.getTerminalType();
                        if (startTerminalType == null) {
                            recorde.setTerminalType(startStudyLog.getTerminalType());
                        }
                        recorde.setId(existRecordeItem.getId());


                    } else {
                        //recorde.setStartTime(new Date(startStudyDateVal - (heartBeatVo.getCount() * 1000)));//开始时间
                        recorde.setStartTime(firstStartDate);

                        recorde.setTerminalDevice("心跳新增的问题数据[" + DateUtil.toShortSeconds(new Date()) + recorde.getOvered() + "]" + startStudyLog.getTerminalDevice() + extedText);

                        String startTerminalType = startStudyLog.getTerminalType();

                        if (startTerminalType != null) {
                            recorde.setTerminalType(startTerminalType);
                        }

                        recorde.setSessionId(startStudyLog.getSessionId());
                        Long studyLogChapterId = startStudyLog.getChapterId();
                        recorde.setCourseId(startStudyLog.getCourseId());
                        recorde.setChapterId(studyLogChapterId);
                        if (studyLogChapterId == null || studyLogChapterId.longValue() == 0) {
                            Optional<StudyLogVo> dataStudyLogChapterIdL = valList.stream().parallel().filter
                                    (studyLogChapterIdL -> (studyLogChapterIdL.getChapterId() != null
                                            && studyLogChapterIdL.getChapterId().longValue() > 0)).findAny();
                            if (dataStudyLogChapterIdL.isPresent()) {
                                StudyLogVo studyLogChapterIdLd = dataStudyLogChapterIdL.get();
                                recorde.setCourseId(studyLogChapterIdLd.getCourseId());
                                recorde.setChapterId(studyLogChapterIdLd.getChapterId());
                            } else {
                                List<StudyLogVo> studyLogVoList = item.getValue();
                                List<StudyLogVo> courseStudyLogVOList = getCourseStudyLogVO(studyLogVoList);
                                errorData.put(sessionId + "章节id为空", courseStudyLogVOList);
                                XxlJobLogger.log("章节id为空的数据={}", JSON.toJSONString(item));
                                continue;
                            }
                        }


                        //把学习记录 加上当前课程的企业id 站点id 部门id
                        String studyAccountId = startStudyLog.getAccountId();
                        if (StringUtils.isBlank(studyAccountId)) {
                            List<StudyLogVo> studyLogVoList = item.getValue();
                            List<StudyLogVo> courseStudyLogVOList = getCourseStudyLogVO(studyLogVoList);
                            errorData.put(sessionId + "账号id为空", courseStudyLogVOList);
                            XxlJobLogger.log("账号id为空的数据={}", JSON.toJSONString(item));
                            continue;
                        }
                        recorde.setAccountId(Long.valueOf(studyAccountId));
                        String studyOrgId = startStudyLog.getOrgId();
                        if (StringUtils.isBlank(studyOrgId) || studyOrgId.equals("0")) {
                            recorde.setOrgId(course.getOrgId());
                        } else {
                            try {
                                recorde.setOrgId(Long.valueOf(studyOrgId));
                            } catch (Exception e) {
                                recorde.setOrgId(course.getOrgId());
                            }
                        }
                        Long studySiteId = startStudyLog.getSiteId();
                        if (Objects.isNull(studySiteId) || studySiteId.longValue() == 0) {
                            recorde.setSiteId(course.getSiteId());
                        } else {
                            recorde.setSiteId(studySiteId);
                        }
                        String studyCompanId = startStudyLog.getCompanyId();
                        if (StringUtils.isBlank(studyCompanId) || studyCompanId.equals("0")) {
                            recorde.setCompanyId(course.getCompanyId());
                        } else {
                            try {
                                recorde.setCompanyId(Long.valueOf(studyCompanId));
                            } catch (NumberFormatException e) {
                                recorde.setCompanyId(course.getCompanyId());
                            }
                        }
                        // 判断章节是否已经被用户学习完成过，如果是，该用户后面的学习都为完成
                        /*int chapterFinished = recordeClient.getCourseChapterFinished(recorde.getAccountId()
                                , recorde.getChapterId());
                        if (chapterFinished > 0) {
                            overed = 1;
                        }*/

                        recorde.setValid(1);
                        recorde.setExitstates(1);

                    }

                    recorde.setProgress(progress);

                    recorde.setOvered(overed); //先设置一次完成状态，如果0，还会第二次赋值

                    int countDuration = heartBeatVo.getCount() * studyLogSize; // 本次学习时长
                    //int countDuration = heartBeatVo.getCount() * groupSize;
                    duration = duration + countDuration;
                    recorde.setDurationTmp(countDuration);// 设置临时的学习时长，解决培训的时长记录问题

                    //long diff =recorde.getEndTime().getTime()- recorde.getStartTime().getTime();
                    recorde.setDuration(duration);// 设置每次学习的学习时长

                    // 计算章节的学习完成状态
                    int dbOvered = 0;
                    if (existRecordeItem != null && existRecordeItem.getOvered().intValue() == 1) {
                        dbOvered = 1;
                    }
                    if (overed == 0 && dbOvered == 0) {
                        // 获取章节的类型
                        Long chapterId = startStudyLog.getChapterId();
                        if (Long.valueOf(0).equals(chapterId)) {
                            chapterId = item.getValue().parallelStream().map(objVal -> objVal.getChapterId() == null ? 0 : objVal.getChapterId().longValue())
                                    .max((a, b) -> a > b ? 1 : -1).get();
                            if (existRecordeItem != null &&
                                    (existRecordeItem.getChapterId() == null || existRecordeItem.getChapterId().longValue() == 0)) {
                                existRecordeItem.setChapterId(chapterId);
                            }
                        }
                        if (chapterId == null || chapterId.intValue() == 0) {
                            if (existRecordeItem != null) {
                                chapterId = existRecordeItem.getChapterId();
                            }
                        }
                        ChapterEntityVo chapter = chapterMap.get(chapterId);
                        if (chapter == null) {
                            //XxlJobLogger.log("{}该章节已经在课程里已被修改，重新查询", chapterId);
                            List<StudyLogVo> studyLogVoList = item.getValue();
                            List<StudyLogVo> courseStudyLogVOList = getCourseStudyLogVO(studyLogVoList);
                            errorData.put(sessionId + "章节不存在", courseStudyLogVOList);
                            XxlJobLogger.log("章节不存在的数据={}", JSON.toJSONString(item));
                            continue;
                        }

                        // 素材类型1音频 2 视频 3 文本 4 富媒体 5 Scrom
                        int materialType = chapter.getMaterialType();
                        switch (materialType) {
                            case 1:
                                int audioComplete = course.getAudioComplete();//音频完成条件 0 播放即完成 1听完整个课程
                                if (audioComplete == 0) {
                                    overed = 1;
                                }
                                break;
                            case 2:
                                int videoComplete = course.getVideoComplete();//视频完成条件 0 播放即完成 1听完整个课程
                                if (videoComplete == 0) {
                                    overed = 1;
                                }
                                break;
                            case 3:
                                int docComplete = course.getDocComplete();//文档完成条件 0 观看即完成 1停留时长达到学习时长即完成
                                overed = getOvered(materialMap, countDuration, overed, chapter, docComplete, //duration 修改成 countDuration（当次学习时长）
                                        recorde.getAccountId(), recorde.getSiteId());
                                break;
                            case 4:
                                int richComplete = course.getRichComplete();//多媒体完成条件 0 观看即完成 1停留时长达到学习时长即完成
                                overed = getOvered(materialMap, countDuration, overed, chapter, richComplete, //duration 修改成 countDuration（当次学习时长）
                                        recorde.getAccountId(), recorde.getSiteId());
                                break;
                            default:
                        }
                        if (overed == -1) {
                            List<StudyLogVo> studyLogVoList = item.getValue();
                            List<StudyLogVo> courseStudyLogVOList = getCourseStudyLogVO(studyLogVoList);
                            errorData.put(sessionId + "素材不存在", courseStudyLogVOList);
                            XxlJobLogger.log("素材不存的数据={}", JSON.toJSONString(item));
                            continue;
                        }
                        recorde.setOvered(overed);

                    }
                    if (existRecordeItem == null) {
                        newRecorde.add(recorde);
                    } else {
                        recorde.setEndTime(lastStudyDate);
                        upRecorde.add(recorde);
                    }
                    XxlJobLogger.log("分组SessionId以后的完成处理{}条数据", excutIndex);
                }

                // 同一门课程的章节学习，获取该课程最后一次章节学习记录
                List<RecordeEntityVo> allRecordes = new ArrayList<>();

                allRecordes.addAll(newRecorde);// 合并学习记录
                allRecordes.addAll(upRecorde);

                Map<String, RecordeEntityVo> upRecordeMap = new HashMap<>();
                if (CollectionUtils.isNotEmpty(upRecorde)) {
                    upRecordeMap = upRecorde.stream().collect(Collectors.toMap(upKey -> upKey.getSessionId(), upVal -> upVal));
                }
                // 发放积分和课程完成状态

                if (CollectionUtils.isNotEmpty(allRecordes)) {
                    XxlJobLogger.log("处理课程完成状态的数据{}条", allRecordes.size());
                    int courseSync = 0;
                    for (RecordeEntityVo sendPoint : allRecordes) {
                        courseSync++;
                        XxlJobLogger.log("处理课程完成状态的数据第{}条", courseSync);
                        String courseLockKey = COURSE_DISTRIBUTED_LOCK + ":" + sendPoint.getCourseId() + "." + sendPoint.getAccountId();
                        String sessionId = sendPoint.getSessionId();
                        int maxRetry = 0;
                        boolean courseExecLockStatus = false;
                        do {
                            courseExecLockStatus = redisDistributedLock.lock(courseLockKey, 5000, 90, 50);
                            maxRetry++;
                        } while (!courseExecLockStatus && maxRetry < 3);
                        try {
                            try {
                                //XxlJobLogger.log("发送处理的数据={}",JSON.toJSONString(sendPoint));
                                courseStudyClient.completionStatus(sendPoint);
                            } catch (Exception e) {
                                XxlJobLogger.log("{}发放积分异常{}", JSON.toJSONString(sendPoint), e);
                                String errKey = sendPoint.getSessionId();
                                List<StudyLogVo> studyLogVoList = data.get(errKey);
                                List<StudyLogVo> courseStudyLogVOList = getCourseStudyLogVO(studyLogVoList);
                                errorData.put(errKey + "发放积分异常", courseStudyLogVOList);
                            }
                            if (upRecordeMap.containsKey(sessionId)) {
                                RecordeEntityVo recordeU = new RecordeEntityVo();
                                recordeU.setId(sendPoint.getId());
                                recordeU.setEndTime(sendPoint.getEndTime());
                                if (Integer.valueOf(1).equals(sendPoint.getOvered())) {
                                    recordeU.setOvered(sendPoint.getOvered());
                                }
                                recordeU.setDuration(sendPoint.getDuration());
                                String deviceInfo = sendPoint.getTerminalDevice();
                                int lengthStr = deviceInfo.length();
                                if (lengthStr < 300) {
                                    recordeU.setTerminalDevice(deviceInfo);
                                } else {
                                    deviceInfo += "->" + sendPoint.getOvered() + "心跳过长" + DateUtil.getSimpleDateFormat("dd日HH:mm:ss").format(new Date());
                                    deviceInfo = deviceInfo.substring(lengthStr - 300);
                                    recordeU.setTerminalDevice(deviceInfo);
                                }
                                /*Date lastTime = sendPoint.getEndTime();
                                Date startTimeTmp = sendPoint.getStartTime();
                                if(lastTime != null && startTimeTmp != null && lastTime.compareTo(startTimeTmp) < 1){
                                    recordeU.setEndTime(new Date((startTimeTmp.getTime()+
                                            recordeU.getDuration().intValue()*1000)));
                                }*/
                                //XxlJobLogger.log("更新数据={}",JSON.toJSONString(recordeU));
                                recordeClient.update(recordeU);
                            } else {
                                recordeClient.save(sendPoint);
                            }
                        } finally {
                            if (courseExecLockStatus) {
                                redisDistributedLock.releaseLock(courseLockKey);
                            }
                        }
                    }
                }

                // 对于存在的学历记录，更新数据
/*                if (CollectionUtils.isNotEmpty(upRecorde)) {
                    XxlJobLogger.log("更新学习记录{}条",upRecorde.size());
                    List<Recorde> customUp = new ArrayList<>();
                    for(Recorde itemUp:upRecorde){
                        Recorde recordeU = new Recorde();
                        recordeU.setId(itemUp.getId());
                        recordeU.setEndTime(itemUp.getEndTime());
                        recordeU.setOvered(itemUp.getOvered());
                        recordeU.setDuration(itemUp.getDuration());
                        customUp.add(recordeU);
                    }
                    boolean retUp = recordeClient.batchUpdate(customUp);
                }
                // 新的学习记录新增
                if (CollectionUtils.isNotEmpty(newRecorde)) {
                    XxlJobLogger.log("新增学习记录{}条",newRecorde.size());
                    recordeClient.saveList(newRecorde);
                }*/

                // 删除本次解析过的学习日志
                boolean flag = studyLogClient.deleteBatch(studyLogs.stream().map(obj -> obj.getId()).collect(Collectors.toList()));
                XxlJobLogger.log("删除本批处studyLog记录状态={}，条数={}", flag, studyLogs.size());
                redisCache.set(REDIS_KEY_MAX, "0");

            } else {
                XxlJobLogger.log("无学习记录");
            }
            if (errorData.size() > 0) {
                try {
                    XxlJobLogger.log("写入错误心跳日志数据");
                    //String fileName = "studyHeatError" + (System.currentTimeMillis()) + ".txt";
                    String fileName = "studyHeatError" + (DateUtil.format(new Date(), "yyyy-MM-dd-HH-mm-ss")) + ".txt";
                    //Files.write(Paths.get("/app/logs/"+fileName), JSON.toJSONString(errorData).getBytes());
                    String lineStr = System.getProperty("line.separator");
                    StringBuilder sb = new StringBuilder();
                    for (Map.Entry<String, List<StudyLogVo>> errItem : errorData.entrySet()) {
                        sb.append(JSON.toJSONString(errItem)).append(lineStr);
                    }
                    Files.write(Paths.get("/app/logs/" + fileName), sb.toString().getBytes());
                } catch (IOException e) {
                    XxlJobLogger.log("写入错误心跳日志数据异常");
                }
            }
            Date endDate = new Date();
            XxlJobLogger.log("心跳处理结束时间=" + DateUtil.format(endDate, "yyyy-MM-dd HH:mm:ss.SSS"));
            XxlJobLogger.log("心跳总耗时{}毫秒", endDate.getTime() - execDate.getTime());
        } finally {
            if (lockStatus) {
                if (redisDistributedLock.releaseLock(REDIS_KEY)) {
                    XxlJobLogger.log(uuid + "批次,解锁成功");
                } else {
                    XxlJobLogger.log(uuid + "批次,解锁失败");
                }
            }
        }
        XxlJobLogger.log(uuid + "批次,心跳处理结束时间{}", DateUtil.format(Calendar.getInstance().getTime(), "yyyy-MM-dd HH:mm:ss.SSS"));
    }

    private List<StudyLogVo> getCourseStudyLogVO(List<StudyLogVo> studyLogVoList) {
        List<StudyLogVo> list = new ArrayList<>();
        if (null != studyLogVoList && studyLogVoList.size() > 0){
            studyLogVoList.forEach(studyLogVo -> {
                StudyLogVo courseVO = new StudyLogVo();
                BeanUtils.copyProperties(studyLogVo,courseVO);
                list.add(courseVO);
            });
        }
        return list;
    }


    private int getOvered(Map<Long, MaterialEntityVo> materialMap, int duration,
                          int overed, ChapterEntityVo chapter, int complete, long accountId, long siteId) {
        if (complete == 0) {
            overed = 1;
        } else {
            // 查询之前学习的时长
            //XxlJobLogger.log("查询历史时长参数为accountId={},siteId={},chapterId="+chapter.getId(),accountId,siteId);
            Long countDuration = recordeClient.countStudyDuration(accountId, siteId, chapter.getId());
            if (Objects.nonNull(countDuration)) {
                duration += countDuration.longValue();
            }
            MaterialEntityVo material = materialMap.get(chapter.getMaterialId());
            if (material == null) {
                return -1;
            }
            int timeLong = 0;
            timeLong = getTimeLong(material, timeLong);
            if (duration >= timeLong) {
                overed = 1;
            }
        }
        return overed;
    }

    private static int getTimeLong(MaterialEntityVo material, int timeLong) {
        Integer textSecond = material.getTextSecond();
        if (Objects.nonNull(textSecond)) {
            timeLong += textSecond.intValue();
        }
        Integer textMinute = material.getTextMinute();
        if (Objects.nonNull(textMinute)) {
            timeLong += textMinute.intValue() * 60;
        }
        Integer textHour = material.getTextHour();
        if (Objects.nonNull(textHour)) {
            timeLong += textHour.intValue() * 3600;
        }
        return timeLong;
    }

    private static void sortList(List<StudyLogVo> valList) {
        valList.sort(new Comparator<StudyLogVo>() {
            @Override
            public int compare(StudyLogVo o1, StudyLogVo o2) {
                /*Long timestamp1 = o1.getTimestamp();
                Long timestamp2 = o2.getTimestamp();*/
                return o1.getNginxTime().compareTo(o2.getNginxTime());
            }
        });
    }

}
